/**
 * EasyBeans
 * Copyright (C) 2007 Bull S.A.S.
 * Contact: easybeans@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: AbsMojo.java 1957 2007-10-15 16:19:40Z benoitf $
 * --------------------------------------------------------------------------
 */

package org.ow2.easybeans.osgi.plugins.launcher;

import java.io.File;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Properties;

import org.apache.maven.artifact.Artifact;
import org.apache.maven.artifact.factory.ArtifactFactory;
import org.apache.maven.artifact.repository.ArtifactRepository;
import org.apache.maven.artifact.resolver.ArtifactNotFoundException;
import org.apache.maven.artifact.resolver.ArtifactResolutionException;
import org.apache.maven.artifact.resolver.ArtifactResolver;
import org.apache.maven.model.Dependency;
import org.apache.maven.model.DependencyManagement;
import org.apache.maven.plugin.AbstractMojo;
import org.apache.maven.plugin.MojoExecutionException;
import org.apache.maven.project.MavenProject;
import org.ow2.easybeans.osgi.plugins.launcher.api.ILauncher;
import org.ow2.easybeans.osgi.plugins.launcher.equinox.EquinoxFramework;
import org.ow2.easybeans.osgi.plugins.launcher.felix.FelixFramework;
import org.ow2.easybeans.osgi.plugins.launcher.knopflerfish.KnopflerfishFramework;
import org.ow2.util.url.URLUtils;

/**
 * Common class for the Mojo (start or stop) and defines the common OSGi stuff.
 * @author Florent BENOIT
 */
public abstract class AbsMojo extends AbstractMojo {

    /**
     * Apache Felix Framework.
     */
    private static final String FELIX_FRAMEWORK = "Felix";

    /**
     * Eclipse Equinox Framework.
     */
    private static final String EQUINOX_FRAMEWORK = "Equinox";

    /**
     * Knopflerfish Framework.
     */
    private static final String KNOPFLERFISH_FRAMEWORK = "Knopflerfish";

    /**
     * Key for keeping the launcher (local or remote).
     */
    private static final String LAUNCHER_KEY = "EasyBeans-OSGi-Launcher";

    /**
     * Path of the build directory.
     * @parameter expression="${project.build.directory}"
     */
    private File buildDirectory;

    /**
     * Artifact Factory used to create Artifacts.
     * @component
     */
    private ArtifactFactory artifactFactory;

    /**
     * Artifact Resolver used to resolve artifacts.
     * @component
     */
    private ArtifactResolver artifactResolver;

    /**
     * The maven project.
     * @parameter expression="${project}"
     * @required
     * @readonly
     */
    private MavenProject project;

    /**
     * Local Maven repository used by the Artifact Resolver.
     * @parameter expression="${localRepository}"
     */
    private ArtifactRepository localRepository;

    /**
     * List of Maven repositories (remote) used by the Artifact Resolver.
     * @parameter expression="${project.remoteArtifactRepositories}"
     */
    private List<?> repositories;

    /**
     * Type of Framework (felix | equinox | knopflerfish).
     * @parameter
     * @required
     */
    private String framework = null;

    /**
     * List of directories that contains bundles.
     * @parameter
     * @optional
     */
    private File[] bundleDirectories = null;

    /**
     * Fork the JVM for the start of the framework ? (default is false)
     * @parameter
     * @optional
     */
    private boolean fork = false;


    /**
     * List of Modules that needs to be installed/started.
     * @parameter
     * @optional
     */
    private Module[] modules;

    /**
     * Configuration file.
     * @parameter
     * @optional
     */
    private File configurationFile = null;

    /**
     * JVM Properties (System properties to set).
     * @parameter
     * @optional
     */
    private Properties jvmProperties;

    /**
     * List of bundles to stop in the given order (by their symbolic name).
     * @parameter
     * @optional
     */
    private String[] stopOrderBundles = null;

    /**
     * List of bundles to start.
     */
    private List<URL> bundles = null;

    /**
     * List of Artifacts to Start.
     */
    private List<Artifact> artifacts = null;

    /**
     * Waiting value after the stop.
     * @parameter
     * @optional
     */
    private long waitAfterStop = 0;

    /**
     * Default constructor.
     */
    public AbsMojo() {
        this.bundles = new LinkedList<URL>();
        this.artifacts = new LinkedList<Artifact>();
    }

    /**
     * Execute the Maven plugin.
     * @throws MojoExecutionException if the file is not generated.
     */
    @SuppressWarnings("unchecked")
    public void execute() throws MojoExecutionException {
        getLog().info("Using " + this.framework + " Framework");

        // Add artifacts given by the user
        resolveArtifacts();
        extractBundlesFromArtifacts();

        // Extract bundles from the given directories
        extractBundlesFromDirectories();

        doExecute();
    }

    /**
     * Subclass execution (code for start, stop, etc).
     * @throws MojoExecutionException if the execution fails
     */
    protected abstract void doExecute() throws MojoExecutionException;

    /**
     * Extract the bundles from the given directories.
     */
    protected void extractBundlesFromDirectories() {
        // No directories, do nothing
        if (bundleDirectories == null) {
            return;
        }

        // Take bundles of the directories
        for (File directory : bundleDirectories) {
            addBundleFromDirectory(directory);
        }

    }

    /**
     * Extract the bundles from the given artifacts.
     */
    protected void extractBundlesFromArtifacts() {
        // Use URL of artifacts
        for (Artifact artifact : artifacts) {
            bundles.add(URLUtils.fileToURL(artifact.getFile()));
        }
    }

    /**
     * Add the bundles that can be found in the given directory.
     * @param directory the directory to scan
     */
    protected void addBundleFromDirectory(final File directory) {
        if (directory.isDirectory()) {
            File[] files = directory.listFiles();
            for (File f : files) {
                if (f.isDirectory()) {
                    addBundleFromDirectory(f);
                } else {
                    bundles.add(URLUtils.fileToURL(f));
                }
            }
        }
    }

    /**
     * Gets a launcher if a previous one is existing or creates a new one if it is permitted.
     * @return a launcher (existing or newly created)
     * @throws MojoExecutionException if no launcher can't be found/created
     */
    @SuppressWarnings("unchecked")
    ILauncher getLauncher() throws MojoExecutionException {
        Map pluginContext = getPluginContext();

        ILauncher launcher = null;

        // Check context ?
        if (pluginContext != null) {
            // Existing launcher ?
            launcher = (ILauncher) pluginContext.get(LAUNCHER_KEY);
            getLog().info("Found a previous launcher in the plugin context");
        }

        if (launcher == null) {
            launcher = createLauncher();
            pluginContext.put(LAUNCHER_KEY, launcher);
        }

        return launcher;
    }


    /**
     * Build a new launcher (local or remote) depending of the fork mode.
     * @return a launcher.
     * @throws MojoExecutionException if no launcher is created
     */
    ILauncher createLauncher() throws MojoExecutionException {
        ILauncher launcher = null;

        // ForK ? create a remote launcher
        if (fork) {
            // Extract URLs of the current classloader

            // Get Method
            ClassLoader cl = Thread.currentThread().getContextClassLoader();
            Method getURLsMethod = null;
            try {
                getURLsMethod = URLClassLoader.class.getDeclaredMethod("getURLs");
            } catch (SecurityException e) {
                throw new MojoExecutionException("Cannot get URLs of the classloader", e);
            } catch (NoSuchMethodException e) {
                throw new MojoExecutionException("Cannot get URLs of the classloader", e);
            }

            // Invoke method
            URL[] urls = null;
            boolean previousAccessible = getURLsMethod.isAccessible();
            getURLsMethod.setAccessible(true);
            try {
                urls = (URL[]) getURLsMethod.invoke(cl);
            } catch (IllegalArgumentException e) {
                throw new MojoExecutionException("Cannot get URLs of the classloader", e);
            } catch (IllegalAccessException e) {
                throw new MojoExecutionException("Cannot get URLs of the classloader", e);
            } catch (InvocationTargetException e) {
                throw new MojoExecutionException("Cannot get URLs of the classloader", e);
            } finally {
                getURLsMethod.setAccessible(previousAccessible);
            }


            launcher = new RemoteLauncher(urls, getConfiguration());
        } else {
            launcher = new LocalLauncher(getConfiguration());
        }

        return launcher;
    }


    /**
     * Return the name of the class for a given framework name.
     * @return the name of the class for a given framework name
     * @throws MojoExecutionException if no framework classname is found
     */
    public String getFrameworkClassname() throws MojoExecutionException {
        // What is the framework ?
        if (FELIX_FRAMEWORK.equalsIgnoreCase(this.framework)) {
            return FelixFramework.class.getName();
        } else if (EQUINOX_FRAMEWORK.equalsIgnoreCase(this.framework)) {
            return EquinoxFramework.class.getName();
        } else if (KNOPFLERFISH_FRAMEWORK.equalsIgnoreCase(this.framework)) {
            return KnopflerfishFramework.class.getName();
        } else {
            throw new MojoExecutionException("No framework was found for the key '" + this.framework + "'.");
        }
    }

    /**
     * Gets the bundles.
     * @return list of bundles
     */
    protected List<URL> getBundles() {
        return bundles;
    }

    /**
     * Found artifacts for the list of the modules specified by the user.
     * @throws MojoExecutionException if artifact resolution fails
     */
    @SuppressWarnings("unchecked")
    protected void resolveArtifacts() throws MojoExecutionException {
        // No module, break here
        if (modules == null) {
            return;
        }

        // for each module, build artifact.
        for (Module module : modules) {
            String groupId = module.getGroupId();
            String artifactId = module.getArtifactId();
            String type = module.getType();
            String version = module.getVersion();

            if (groupId == null) {
                throw new MojoExecutionException("No groupdId set for module '" + module + "'");
            }
            if (artifactId == null) {
                throw new MojoExecutionException("No artifactId set for module '" + module + "'");
            }

            if (type == null) {
                getLog().debug("No type set for module '" + module + "', assume it is jar type");
                type = "jar";
                module.setType(type);
            }

            // No version
            if (version == null) {
                // try to find version with dependency management
                DependencyManagement dependencyManagement = project.getDependencyManagement();

                if (dependencyManagement != null) {
                    // try to find a version for this component
                    List<Dependency> dependencies = dependencyManagement.getDependencies();
                    if (dependencies != null) {
                        for (Dependency dependency : dependencies) {
                            getLog().debug("Dependency = " + dependency);
                            if (groupId.equals(dependency.getGroupId())
                                    && artifactId.equals(dependency.getArtifactId())
                                    && type.equals(dependency.getType())) {
                                getLog().debug("Using version of dependency management" + dependency.getVersion());
                                version = dependency.getVersion();
                                break;
                            }
                        }
                    }
                }
                if (version == null) {
                    throw new MojoExecutionException("No version set for module '" + module + "'");
                }
            }

            Artifact artifact = artifactFactory.createBuildArtifact(groupId, artifactId, version, type);
            getLog().debug("Built Artifact = " + artifact);

            // now resolve this artifact
            try {
                artifactResolver.resolve(artifact, repositories, localRepository);
            } catch (ArtifactResolutionException e) {
                throw new MojoExecutionException("Cannot resolve artifact '" + artifactId + "'.", e);
            } catch (ArtifactNotFoundException e) {
                throw new MojoExecutionException("Artifact '" + artifactId + "' not found.", e);
            }

            // Add artifact
            artifacts.add(artifact);
        }
    }

    /**
     * Gets the configuration used by the launcher/manager.
     * @return a configuration object
     * @throws MojoExecutionException if the configuration can't be built.
     */
    protected Configuration getConfiguration() throws MojoExecutionException {
        Configuration configuration = new Configuration();

        // Time to wait after the stop
        configuration.setWaitAfterStop(waitAfterStop);

        // Stop bundles in a given order
        configuration.setStopOrderBundles(stopOrderBundles);

        // JVM properties
        configuration.setSystemProperties(jvmProperties);

        // Name of the framework to instantiate
        configuration.setFrameworkClassname(getFrameworkClassname());

        // Bundles to deploy
        configuration.setBundles(bundles);

        // Set working directory
        configuration.setWorkingDirectory(new File(buildDirectory, this.framework));

        // Config file URL
        if (configurationFile != null) {
            configuration.setURLConfigFile(URLUtils.fileToURL(configurationFile));
        }

        return configuration;
    }

}
