/**
 * EasyBeans
 * Copyright (C) 2007 Bull S.A.S.
 * Contact: easybeans@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: Bootstrap.java 1941 2007-10-14 17:57:14Z benoitf $
 * --------------------------------------------------------------------------
 */

package org.ow2.easybeans.osgi.plugins.launcher;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.util.Enumeration;
import java.util.Properties;

import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;

/**
 * @author Florent BENOIT
 */
public final class Bootstrap {

    /**
     * Seconds is 1000 millis.
     */
    private static final long SECONDS = 1000L;

    /**
     * One minute is 60 seconds.
     */
    private static final long ONE_MINUTE = 60 * SECONDS;

    /**
     * Pause while analyzing the monitoring file.
     */
    private static final long WAIT_MONITORING = 500L;

    /**
     * Name of the configuration argument.
     */
    public static final String CONFIGURATION_PATH = "bootstrap.configuration.path";

    /**
     * File that is used to receive calls for the shutdown.
     */
    public static final String MONITORED_FILE = System.getProperty("java.io.tmpdir") + File.separator
            + System.getProperty("user.name") + "Launcher.lock";

    /**
     * Started instruction.
     */
    public static final String STARTED_INSTRUCTION = "STARTED";

    /**
     * Stopped instruction.
     */
    public static final String STOPPED_INSTRUCTION = "STOPPED";

    /**
     * Logger.
     */
    private static Log logger = LogFactory.getLog(Bootstrap.class);

    /**
     * Configuration.
     */
    private Configuration configuration = null;

    /**
     * Framework manager that will be used to start/stop framework.
     */
    private FrameworkManager frameworkManager = null;

    /**
     * Already called shutdown method ? if yes, avoid to do it again.
     */
    private boolean shutdownDone = false;

    /**
     * Utility class.
     */
    private Bootstrap() {

    }

    /**
     * Starts the Framework and launch threads for monitor a shutdown call.
     * @param args the arguments with the path to the conf file.
     */
    @SuppressWarnings("unchecked")
    public void start(final String[] args) {

        // Timeout of the JVM
        new Thread() {

            public void run() {
                try {
                    Thread.sleep(ONE_MINUTE);
                } catch (InterruptedException e) {
                    logger.error("Cannot sleep");
                }

                // Exit JVM after 1minute
                System.exit(0);
            }
        }.start();

        // Read arguments in order to know where is configuration object
        if (args == null || args.length != 2) {
            logger.error("Invalid arguments");
            return;

        }

        // Get path
        String configurationPath = args[1];
        if (configurationPath == null) {
            logger.error("No configuration path found");
            return;
        }

        logger.debug("Using configuration stored in {0}", configurationPath);

        // Read configuration object from the filesystem
        File file = new File(configurationPath);
        FileInputStream fileInputStream = null;
        ObjectInputStream objectInputStream = null;
        Object readObject = null;
        try {
            fileInputStream = new FileInputStream(file);
            objectInputStream = new ObjectInputStream(fileInputStream);
            readObject = objectInputStream.readObject();
        } catch (FileNotFoundException e) {
            logger.error("Cannot find the file {0}", file, e);
            return;
        } catch (IOException e) {
            logger.error("Cannot read the file {0}", file, e);
            return;
        } catch (ClassNotFoundException e) {
            logger.error("Cannot find the Configuration class", e);
            return;
        } finally {
            // Close everything
            if (fileInputStream != null) {
                try {
                    fileInputStream.close();
                } catch (IOException e) {
                    logger.error("Cannot close stream", e);
                }
            }
            if (objectInputStream != null) {
                try {
                    objectInputStream.close();
                } catch (IOException e) {
                    logger.error("Cannot close stream", e);
                }
            }
        }

        // Ensure that object is a valid type
        if (readObject instanceof Configuration) {
            this.configuration = (Configuration) readObject;
        } else {
            logger.error("Object read on the disk is not a Configuration object. Found {0}", readObject);
            return;
        }

        // Set system properties
        Properties jvmProperties = configuration.getSystemProperties();
        if (jvmProperties != null) {
            Enumeration<Object> enumeration = jvmProperties.keys();
            while (enumeration.hasMoreElements()) {
                String key = enumeration.nextElement().toString();
                String value = jvmProperties.get(key).toString();
                // property already exists ?
                String oldValue = System.getProperty(key);
                if (!value.equals(oldValue)) {
                    System.setProperty(key, value);
                    logger.info("Adding System property '" + key + "' with value '" + value + "'");
                }
            }
        }

        try {
            this.frameworkManager = FrameworkManagerHelper.getFrameworkManager(configuration);
        } catch (FrameworkManagerException e) {
            logger.error("Cannot get the framework manager", e);
            return;
        }

        // start it
        try {
            this.frameworkManager.start();
        } catch (FrameworkManagerException e) {
            logger.error("Cannot start the framework", e);
            return;
        }

        // Launch a thread that monitor a file and stop the server if it
        // contains STOPPED_INSTRUCTION

        new Thread() {

            public void run() {
                boolean stopSend = false;
                while (!stopSend) {
                    File file = new File(MONITORED_FILE);
                    // File exists, read content
                    if (file.exists()) {
                        FileReader fileReader;
                        try {
                            fileReader = new FileReader(file);
                            BufferedReader bufferedReader = new BufferedReader(fileReader);
                            String instruction = bufferedReader.readLine();
                            if (STOPPED_INSTRUCTION.equals(instruction)) {
                                logger.info("Detect STOP instruction !");
                                stopSend = true;
                                shutdown();
                            }
                        } catch (FileNotFoundException e) {
                            logger.debug("File not found", e);
                        } catch (IOException e) {
                            logger.debug("Cannot read file", e);
                        }
                    }

                    // Wait a little
                    try {
                        Thread.sleep(WAIT_MONITORING);
                    } catch (InterruptedException e) {
                        // TODO Auto-generated catch block
                        e.printStackTrace();
                    }
                }
            }
        }.start();

        logger.error("Waiting...");
        try {
            Thread.sleep(ONE_MINUTE);
        } catch (InterruptedException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
        logger.error("Stopping after timeout...");

        shutdown();
    }


    /**
     * Stop this framework.
     */
    public synchronized void shutdown() {
        if (shutdownDone) {
            return;
        }
        shutdownDone = true;
        try {
            this.frameworkManager.stop();
        } catch (FrameworkManagerException e) {
            logger.error("Cannot stop the framework", e);
        }

        logger.info("Waiting {0} seconds before killing the JVM", configuration.getWaitAfterStop());
        try {
            Thread.sleep(configuration.getWaitAfterStop() * SECONDS);
        } catch (InterruptedException e) {
            logger.error("Cannot wait", e);
        }

        System.exit(0);
    }

    /**
     * Starts the bootstrap class. It will read the configuration from a given
     * argument.
     * @param args contains the path to the configuration object to use.
     */
    public static void main(final String[] args) {
        new Bootstrap().start(args);
    }
}
