/**
 * EasyBeans
 * Copyright (C) 2007 Bull S.A.S.
 * Contact: easybeans@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: FrameworkManager.java 1957 2007-10-15 16:19:40Z benoitf $
 * --------------------------------------------------------------------------
 */

package org.ow2.easybeans.osgi.plugins.launcher;

import java.net.URL;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

import org.osgi.framework.Bundle;
import org.osgi.framework.BundleContext;
import org.osgi.framework.BundleException;
import org.ow2.easybeans.osgi.plugins.launcher.api.FrameworkException;
import org.ow2.easybeans.osgi.plugins.launcher.api.IFramework;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;

/**
 * @author Florent BENOIT
 */
public class FrameworkManager {

    /**
     * Seconds is 1000 millis.
     */
    private static final long SECONDS = 1000L;

    /**
     * Logger.
     */
    private Log logger = LogFactory.getLog(FrameworkManager.class);

    /**
     * Configuration object.
     */
    private Configuration configuration = null;

    /**
     * Framework to use.
     */
    private IFramework framework;

    /**
     * Execute the selected framework.
     * @throws FrameworkManagerException if plugin is not executed
     */
    public void start() throws FrameworkManagerException {

        // Set working directory
        framework.setWorkingDirectory(configuration.getWorkingDirectory());

        // Configure Framework
        try {
            framework.configure(configuration.getURLConfigFile());
        } catch (FrameworkException e) {
            throw new FrameworkManagerException("Cannot configure the framework", e);
        }

        // start the framework
        logger.info("Starting Framework");
        try {
            framework.start();
        } catch (FrameworkException e) {
            throw new FrameworkManagerException("Cannot start the framework", e);
        }

        // Deploy the bundles given by the user
        // Use the Bundle context of the framework
        BundleContext bundleContext = framework.getBundleContext();

        logger.info("Installing bundles");
        List<Bundle> osgiBundles = new LinkedList<Bundle>();
        // Install bundles
        for (URL bundle : configuration.getBundles()) {
            try {
                logger.info(" -> Installing bundle {0}", bundle);
                Bundle osgiBundle = bundleContext.installBundle(bundle.toExternalForm());
                osgiBundles.add(osgiBundle);
                logger.info("     -> Bundle is {0}", osgiBundle);
            } catch (BundleException e) {
                throw new FrameworkManagerException("Cannot install the bundle '" + bundle + "'", e);
            }
        }

        // Compute list of bundles to delayed.
        List<String> patternsDelayed = new ArrayList<String>();
        if (configuration.getDelayedBundles() != null) {
            for (DelayedBundlePattern delayedBundlePattern : configuration.getDelayedBundles()) {
                logger.info("Adding {0}", delayedBundlePattern);
                patternsDelayed.add(delayedBundlePattern.getPattern());
            }
        }

        // Auto Start bundles in a specified order ?
        String[] orderedBundles = configuration.getStartOrderBundles();
        if (orderedBundles != null) {
            // Search the bundle of the installed bundle and if there is a
            // match, start it
            for (String symbolicName : orderedBundles) {
                for (Bundle bundle : osgiBundles) {
                    if (symbolicName.equals(bundle.getSymbolicName())) {
                        logger.info(" -> Starting ordered bundle {0}", bundle);
                        try {
                            bundle.start();
                        } catch (BundleException e) {
                            throw new FrameworkManagerException("Cannot start the bundle '" + bundle + "'", e);
                        }
                    }
                }
            }
        }

        // Start remaining bundles
        logger.info("Starting bundles");
        for (Bundle bundle : osgiBundles) {
            // Delayed ? wait
            boolean delayed = false;
            // Delayed bundle ?
            for (String pattern : patternsDelayed) {
                if (bundle.getSymbolicName().matches(pattern)) {
                    logger.info("  -> Delaying start of {0}", bundle.getSymbolicName());
                    delayed = true;
                    break;
                }
            }

            // Start only if not delayed
            if (!delayed) {
                try {
                    logger.info(" -> Starting bundle {0}", bundle);
                    bundle.start();
                } catch (BundleException e) {
                    throw new FrameworkManagerException("Cannot start the bundle '" + bundle + "'", e);
                }
            }
        }

        // Start bundle after their delayed time.
        // TODO : Launch in a thread...
        for (DelayedBundlePattern delayedBundle : configuration.getDelayedBundles()) {
            // Wait given time
            try {
                Thread.sleep(delayedBundle.getSecondsToWait() * SECONDS);
            } catch (InterruptedException e) {
                throw new FrameworkManagerException("Cannot wait for delayed bundles", e);
            }

            String pattern = delayedBundle.getPattern();

            // Start all matching bundles
            for (Bundle bundle : osgiBundles) {
                if (bundle.getSymbolicName().matches(pattern)) {
                    try {
                        logger.info(" -> Starting delayed bundle {0}", bundle);
                        bundle.start();
                    } catch (BundleException e) {
                        throw new FrameworkManagerException("Cannot start the bundle '" + bundle + "'", e);
                    }
                }
            }
        }

        if (configuration.getWaitAfterStart() > 0) {
            try {
                Thread.sleep(configuration.getWaitAfterStart() * SECONDS);
            } catch (InterruptedException e) {
                throw new FrameworkManagerException("Cannot wait", e);
            }
        }

    }

    /**
     * Execute the plugin in the sub class.
     * @throws FrameworkManagerException if the framework can't be stopped
     */
    public void stop() throws FrameworkManagerException {

        // Auto stop bundles in a specified order ?
        String[] orderedBundles = configuration.getStopOrderBundles();
        if (orderedBundles != null) {

            // List of bundles managed by the framework
            Bundle[] frameworkBundles = framework.getBundleContext().getBundles();
            if (frameworkBundles != null) {

                // Search the bundle of the installed bundle and if there is a
                // match, stop it
                for (String symbolicName : orderedBundles) {

                    for (Bundle bundle : frameworkBundles) {
                        if (symbolicName.equals(bundle.getSymbolicName())) {
                            logger.info(" -> Stopping ordered bundle {0}", bundle);
                            try {
                                bundle.stop();
                            } catch (BundleException e) {
                                throw new FrameworkManagerException("Cannot stop the bundle '" + bundle + "'", e);
                            }
                        }
                    }
                }
            }
        }

        logger.info("Stopping Framework...");
        // Stop the framework
        try {
            framework.stop();
        } catch (FrameworkException e) {
            throw new FrameworkManagerException("Cannot stop the framework", e);
        }

        logger.info("Framework stopped");
    }

    /**
     * Gets the configuration used by this manager.
     * @return the configuration.
     */
    public Configuration getConfiguration() {
        return configuration;
    }

    /**
     * Sets the configuration used by this manager.
     * @param configuration the given configuration.
     */
    public void setConfiguration(final Configuration configuration) {
        this.configuration = configuration;
    }

    /**
     * Sets the framework that the manager will use to call lifecycle methods.
     * @param framework the given framework
     */
    public void setFramework(final IFramework framework) {
        this.framework = framework;
    }

}
